/*
 * (c) 2006 Thomas Maier-Komor
 * LICENSE: GPLv2, see file COPYING for details
 */

#include <assert.h>
#include <errno.h>
#include <fcntl.h>
#include <limits.h>
#include <signal.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <unistd.h>

#include "log.h"
#include "tokens.h"


typedef struct arg_s {
	char *arg;
	struct arg_s *prev;
} arg_t;

typedef struct job_s {
	pid_t pid;
	char **args;
	size_t numargs;
	int in, out, err;
	struct job_s *next;
} job_t;

static arg_t *Baseargs = 0;
static size_t Numbaseargs = 0, Running = 0, Waiting = 0;
static unsigned long Limit = 0;
static job_t *Executing = 0, *Ready = 0, *ReadyEnd = 0;
static int Null, Nullin;
static char *Path;

extern int yylex(void);
extern verbose_t Verbose;


static void *Malloc(size_t s)
{
	int tries = 0, time = 5;
	for (;;) {
		void *r = malloc(s);
		if (r != 0)
			return r;
		warn("out of memory - retrying in %d seconds\n",time);
		(void) sleep(time);
		time += 5;
		++tries;
		if (tries == 10) {
			warn("Persisiting low memory situation: stopping.\n"
				"Please send SIGCONT to contiue.\n");
			if (-1 == raise(SIGSTOP))
				abort();
			time = 5;
			tries = 0;
		}
	}
}


static void exec_job(job_t *j)
{
	pid_t pid;
	pid = fork();
	if (pid == -1)
		error("error could not fork: %s\n",strerror(errno));
	if (pid != 0) {
		char **a = j->args;
		j->pid = pid;
		assert(a);
		info("[%6d]: ",pid);
		info("%s",*a++);
		while(*a)
			info(" %s",*a++);
		info("\n");
		j->next = Executing;
		if ((j->in != -1) && (j->in != Nullin))
			(void) close(j->in);
		if ((j->out != -1) && (j->in != Null))
			(void) close(j->out);
		if ((j->err != -1) && (j->in != Null))
			(void) close(j->err);
		Executing = j;
		if (--Waiting)
			dbug("%d jobs still waiting in the queue\n",Waiting);
		++Running;
	} else {
		int err;
		if (j->in != -1) {
			(void) close(STDIN_FILENO);
			if (-1 == dup2(j->in,STDIN_FILENO))
				warn("unable to redirect stdin: %s\n",strerror(errno));
		}
		if (j->out != -1) {
			(void) close(STDOUT_FILENO);
			if (-1 == dup2(j->out,STDOUT_FILENO))
				warn("unable to redirect stdout: %s\n",strerror(errno));
		}
		if (j->err != -1) {
			(void) close(STDERR_FILENO);
			if (-1 == dup2(j->err,STDERR_FILENO))
				warn("unable to redirect stderr: %s\n",strerror(errno));
		}
		err = execv(j->args[0],j->args);
		assert(err == -1);
		error("failed to execute %s: %s\n",j->args[0],strerror(errno));
		exit(EXIT_FAILURE);
	}
}


static int resolve_symlink(char *cmd)
{
	char lt[PATH_MAX+1];
	/* handle symlinks */
	bzero(lt,sizeof(lt));
	while (-1 != readlink(cmd,lt,PATH_MAX)) {
		dbug("is symlink: %s\n",lt);
		/* handle relative links */
		if (lt[0] != '/') { 
			char *base = strrchr(cmd,'/');   
			assert((base != 0) && (base >= cmd));
			++base;
			if (strlen(lt) + (base - cmd) > PATH_MAX) {
				warn("command exceed PATH_MAX limit of your system\n");
				return -1;
			}
			(void) strcpy(base,lt);
			dbug("resolved to: %s\n",cmd);
		} else
			(void) strcpy(cmd,lt);
	}
	return 0;
}


static char *complete_exe(char *exe)
{
	char *token, *next;

	if ((exe[0] == '/') && (0 == access(exe,X_OK)))
                return exe;
	token = Path;
	next = strchr(Path,':');
        while (token && *token) {
		char cmd[PATH_MAX+1];
		size_t off;

		if (next)
			*next = 0;
		if ((off = strlen(token)) > PATH_MAX) {
			warn("command exceed PATH_MAX limit of your system\n");
			return 0;
		}
		(void) strcpy(cmd, token);
		if (cmd[off-1] != '/') {
			cmd[off] = '/';
			++off;
		}
		if (strlen(exe) + off > PATH_MAX) {
			warn("command exceed PATH_MAX limit of your system\n");
			return 0;
		}
		(void) strcpy(cmd+off, exe);
		if (next) {
			*next = ':';
			token = next + 1;
			next = strchr(token,':');
		} else
			token = 0;
		if (-1 == access(cmd,X_OK))
			continue;
		dbug("candidate: %s\n",cmd);
		if (resolve_symlink(cmd))
			return 0;
		if (0 == access(cmd,X_OK)) {
			dbug("using: %s\n",cmd);
			free(exe);
                        return strdup(cmd);
		}
        }
        return 0;
}


static void start_jobs(void)
{
	while ((Running < Limit) && Ready) {
		job_t *j = Ready;
		Ready = j->next;
		if (Ready == 0)
			ReadyEnd = 0;
		exec_job(j);
	}
}


static void add_job(arg_t *a, size_t n, int in, int out, int err)
{
	char **argv, *exe;
	int i = n + Numbaseargs + 1;
	job_t *j;

	assert(a);
	assert(n > 0);
	j = Malloc(sizeof(job_t));
	bzero(j,sizeof(job_t));
	argv = Malloc(sizeof(char *) * i);
	j->in = in;
	j->out = out;
	j->err = err;
	j->args = argv;
	--i;
	argv[i] = 0;
	j->numargs = i;
	while (n) {
		arg_t *c = a;
		assert(a);
		a = c->prev;
		--n;
		--i;
		argv[i] = c->arg;
		free(c);
	}
	assert(a == 0);
	a = Baseargs;
	while (i) {
		arg_t *c = a;
		assert(a);
		a = c->prev;
		--i;
		argv[i] = strdup(c->arg);
	}
	assert(a == 0);
	exe = complete_exe(argv[0]);
	if (0 == exe) {
		int i;
		warn("could not find executable in path for %s\n",argv[0]);
		for (i = 0; i < j->numargs; ++i)
			free(argv[i]);
		free(argv);
		free(j);
		return;
	}
	argv[0] = exe;
	while (*argv) {
		dbug("argument %d: <%s>\n",argv - j->args, *argv);
		++argv;
	}
	j->next = 0;
	if (ReadyEnd)
		ReadyEnd->next = j;
	else
		Ready = j;
	ReadyEnd = j;
	++Waiting;
	start_jobs();
	if (Ready)
		dbug("job queued\n");
}


static void clear_job(pid_t pid, int ret)
{
	int i;
	job_t *j = Executing, *lj = 0;

	while (j && (j->pid != pid)) {
		lj = j;
		j = j->next;
	}
	assert(j);
	if (lj)
		lj->next = j->next;
	else
		Executing = j->next;
	if (ret == 0) 
		info("[%6d]= 0\n",pid);
	else {
		char **a = j->args;
		warn("[%6d/\"",pid);
		warn("%s",*a);
		++a;
		while (*a) {
			warn(" %s",*a);
			++a;
		}
		warn("\"]= %d\n",ret);
	}
	for (i = 0; i < j->numargs; ++i)
		free(j->args[i]);
	free(j->args);
	free(j);
	--Running;
}


static void usage(void)
{
	(void) printf(
	"xjobs [options] [command [argument ...]]\n"
	"========================================\n"
	"valid options are:\n"
	"-h           : print this usage information and exit\n"
	"-j <num>     : maximum number of jobs to execute in parallel\n"
	"-s <file>    : script to execute (default: read from stdin)\n"
	"-n           : redirect stdout and stderr of childs to /dev/null\n"
	"-v <level>   : set verbosity to level\n"
	"               (0=silent,1=error,2=warning,3=info,4=debug)\n"
	"-V           : print version and exit\n"
	);
	exit(0);
}


static void version(void)
{
	(void) printf(
	"xjobs version 20060409\n"
	"Copyright 2006, Thomas Maier-Komor\n"
	"License: GPLv2\n"
	);
	exit(0);
}


int main(int argc, char **argv)
{
	typedef enum { init, std, white, dquoted, squoted, desc, sesc } state_t;
	int i, jobin = -1, jobout = -1, joberr = -1;
	extern char *optarg;
	extern int optind, optopt;
	arg_t *args = 0;
	size_t numargs = 0;
	token_t t;

	Path = getenv("PATH");
	while ((i = getopt(argc,argv,"hj:ns:v:V")) != -1) {
		switch (i) {
		default:
			abort();
			break;
		case 'h':
			usage();
			break;
		case 'j':
			if (1 == sscanf(optarg,"%lu",&Limit))
				dbug("maximum number of jobs set to %d\n",Limit);
			break;
		case 'n':
			Null = open("/dev/null",O_WRONLY);
			if (Null == -1)
				error("could not open /dev/null: %s\n",strerror(errno));
			else
				dbug("output redirected to /dev/null\n");
			jobout = joberr = Null;
			break;
		case 's':
			{
				extern FILE *yyin;
				FILE *in = fopen(optarg,"r");
				if (in == 0)
					error("could not open input file %s: %s\n",optarg,strerror(errno));
				yyin = in;
				dbug("input file set to %s\n",optarg);
				break;
			}
		case 'v':
			if (1 == sscanf(optarg,"%ld",&Verbose)) {
				if (Verbose < Silent || Verbose > Debug) {
					error("value out of range for option -v\n");
					Verbose = 0;
				}
				dbug("verbose set to %d\n",Verbose);
			} else
				Verbose = 0;
			break;
		case 'V':
			version();
			break;
		case '?':
			error("unkonwn option -%c\n",optopt);
			break;
		case ':':
			error("option -%c requires an operand\n",optopt);
			break;
		}
	}
        dbug("PATH=\"%s\"\n", Path);
	if (Limit == 0) {
		Limit = sysconf(_SC_NPROCESSORS_ONLN);
		dbug("%d processors currently online (default job limit)\n",Limit);
	}
	jobin = Nullin = open("/dev/null",O_RDONLY);
	if (Nullin == -1)
		error("could not open /dev/null: %s\n",strerror(errno));

	while (optind < argc) {
		arg_t *a = Malloc(sizeof(arg_t));
		a->arg = argv[optind];
		if ((0 == strcmp(">",argv[optind]))
			|| (0 == strcmp("<",argv[optind]))
			|| (0 == strcmp(">&",argv[optind]))) {
			warn("I/O redirection not supported as command line parameter\n");
			optind += 2;
		} else {
			a->prev = Baseargs;
			Baseargs = a;
			++Numbaseargs;
			++optind;
		}
	}

	while (t = yylex()) {
		extern const char *Buffer;
		pid_t pid;
		int ret;

		switch (t) {
		default:
			abort();
			break;
		case INPUT:
			t = yylex();
			if ((t != UQUOTED) && (t != DQUOTED) && (t != SQUOTED)) {
				warn("invalid argument for input redirector <");
				do t = yylex(); while (t && (t != EOL));
			} else {
				jobin = open(Buffer,O_RDONLY);
				if (jobin == -1)
					warn("could not open input file: %s\n",strerror(errno));
				else
					dbug("job input redirected to %s\n",Buffer);
			}
			break;
		case OUTPUT:
			t = yylex();
			if ((t != UQUOTED) && (t != DQUOTED) && (t != SQUOTED)) {
				warn("invalid argument for output redirector >");
				do t = yylex(); while (t && (t != EOL));
			} else {
				jobout = open(Buffer,O_WRONLY|O_CREAT,0666);
				if (jobout == -1)
					warn("could not open output file: %s\n",strerror(errno));
				else
					dbug("job output redirected to %s\n",Buffer);
			}
			break;
		case ERRLOG:
			t = yylex();
			if ((t != UQUOTED) && (t != DQUOTED) && (t != SQUOTED)) {
				warn("invalid argument for log redirector >&");
				do t = yylex(); while (t && (t != EOL));
			} else {
				joberr = jobout = open(Buffer,O_WRONLY|O_CREAT,0666);
				if (joberr == -1)
					warn("could not open log file: %s\n",strerror(errno));
				else
					dbug("job errors redirected to %s\n",Buffer);
			}
			break;
		case UQUOTED:
		case DQUOTED:
		case SQUOTED:
			{
				arg_t *a = Malloc(sizeof(arg_t));
				bzero(a,sizeof(arg_t));
				a->arg = strdup(Buffer);
				a->prev = args;
				args = a;
				++numargs;
				break;
			}
		case EOL:
			if (numargs)
				add_job(args,numargs,jobin,jobout,joberr);
			args = 0;
			numargs = 0;
			jobin = Nullin;
			jobout = joberr = Null;
			break;
		case ERR:
			info("ignoring erroneous line\n");
			break;
		}
		
		if (Executing && (pid = waitpid(0,&ret,WNOHANG))) {
			if (pid != -1) {
				clear_job(pid,ret);
			}
		}
		start_jobs();
	}

	while (Executing || Ready) {
		pid_t pid;
		int ret;
		start_jobs();
		pid = waitpid(0,&ret,0);
		clear_job(pid,ret);
	}
	return 0;
}

